from typing import Any, Optional

from fastapi import FastAPI, status
from fastapi.responses import ORJSONResponse
from fastapi.exceptions import HTTPException
from pydantic import BaseModel, Field

from .utils import ObjectCollection, api_object_parser
from ..settings import MODEL_MAP
from ..ranker import ReRankPipeline

server = FastAPI()
pipeline_map: dict[str, ReRankPipeline] = {}

def get_pipeline(__id: str):
    if pipeline_map.get(__id) is None:
        pipeline_map[__id] = ReRankPipeline.from_model_id(__id)
    return pipeline_map[__id]


class SchemaContext(BaseModel):
    pre: Optional[str] = Field(None, description="schema for pre-processing input.")
    ctx: Optional[str] = Field(None, description="schema for extracting context.")
    post: Optional[str] = Field(None, description="schema for extracting field after reranking.")
    
class RerankContext(BaseModel):
    model: str = Field("ms-marco-TinyBERT-L-2-v2", description="model to use for reranking.")
    contexts: ObjectCollection = Field(..., description="contexts to rerank.")
    query: str = Field(..., description="query for reranking evaluation.")
    threshold: Optional[float] = Field(None, ge=0.0, le=1.0, description="filter contexts using threshold.")
    map_score: bool = Field(False, description="map relevance score with context")
    schema_: Optional[SchemaContext] = Field(default=None, alias='schema')


@server.get('/models', response_class=ORJSONResponse)
def list_models():
    return list(MODEL_MAP.keys())

@server.post('/rerank')
def rerank_endpoint(ctx: RerankContext):
    if not ctx.contexts:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="contexts field cannot be an empty array or object"
        )

    if ctx.model not in MODEL_MAP:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"{ctx.model!r} model is not available"
        )
    
    schema = ctx.schema_ or SchemaContext()
    if schema.pre is not None:
        contexts = api_object_parser(ctx.contexts, schema=schema.pre)
        if isinstance(contexts, list) and not contexts:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Empty array after pre-processing"
            )
        no_list_err = "Pre-processing must result into an array of objects"

    else:
        contexts = ctx.contexts
        no_list_err = "Expected an array of string or object. 'pre' schema might help"

    if not isinstance(contexts, list):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=no_list_err
        )

    ctx_schema = schema.ctx or '.'
    post_schema = schema.post or '.'
    pipeline = get_pipeline(ctx.model)
    try:
        if ctx.map_score is False:
            reranked = pipeline.invoke(
                query=ctx.query, 
                contexts=contexts, 
                threshold=ctx.threshold,
                key=lambda x: api_object_parser(x, ctx_schema)
            )
            
            return [api_object_parser(context, post_schema) for context in reranked]
        else:
            reranked_tup = pipeline.invoke_with_score(
                query=ctx.query, 
                contexts=contexts, 
                threshold=ctx.threshold,
                key=lambda x: api_object_parser(x, ctx_schema)
            )

            return [
                {'score': score, 'context': api_object_parser(context, post_schema)} 
                for (score, context) in reranked_tup
            ]
    except TypeError:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail='Context processing must result into string'
        )
    
def _serve(host: str, port: int):
    import uvicorn
    try:
        uvicorn.run(server, host=host, port=port)
    except KeyboardInterrupt:
        exit(0)
    
if __name__ == "__main__":
    _serve(host='0.0.0.0', port=12345)


import sys
from typing import TypeAlias, Any


ObjectCollection: TypeAlias = dict[str, Any] | list[Any]
ObjectScalar: TypeAlias = bool | float | int | str
ObjectValue: TypeAlias = ObjectCollection | ObjectScalar    

def object_parser(obj: ObjectValue, schema: str) -> ObjectValue:
    import re
    from itertools import groupby

    if schema == '.':
        return obj
    
    usable_schema = '.['.join(re.split(r'(?:\[|\.\[)', schema))
    if not re.match(
        pattern=r"^(?:(?:[.](?:[\w]+|\[\d?\]))+)$", string=usable_schema
    ):
        raise ValueError(f'{schema!r} is not a valid schema.')

    def __inner__(_in: ObjectValue, keys: list[str]):
        for idx, key in enumerate(keys):
            _match = re.search(r'^\[(\d)?\]$', key)
            try:
                if _match is None:
                    _in = _in[key]
                    continue

                obj_idx = _match.group(1)
                if obj_idx is not None:
                    _in = _in[int(obj_idx)]
                    continue
            except (KeyError, IndexError):
                raise ValueError(f'{schema!r} schema not compatible with input data.')

            _keys = keys[idx + 1:]
            if not _keys:
                return _in

            return [__inner__(item, _keys) for item in _in][0]
        return _in
    return __inner__(
        obj, [k for k, _ in groupby(usable_schema.lstrip('.').split('.'))]
    )

def read_stdin(readlines: bool = False):
    """Read values from standard input (stdin). """
    if sys.stdin.isatty():
        return
    try:
        if readlines is False:
            return sys.stdin.read().rstrip('\n')
        return [_.strip('\n') for _ in sys.stdin if _]
    except KeyboardInterrupt:
        return
    
def print_and_exit(msg: str, code: int = 0):
    stream = sys.stdout if not code else sys.stderr
    print(msg, file=stream)
    exit(code)

def cli_object_parser(obj: ObjectValue, schema: str):
    try:
        return object_parser(obj=obj, schema=schema)
    except ValueError as e:
        print_and_exit(e.args[0], code=1)

def api_object_parser(obj: ObjectValue, schema: str):
    from fastapi import status, HTTPException
    try:
        return object_parser(obj=obj, schema=schema)
    except ValueError as e:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, 
            detail=e.args[0]
        )


import json
from pathlib import Path
from collections import OrderedDict
from typing import (
    overload, cast, Any, Optional, Iterable, Callable, TypeVar
)

import numpy as np
import onnxruntime as ort
from tokenizers import AddedToken, Tokenizer as TokenizerLoader

from . import settings


_T = TypeVar("_T")


class Ranker:
    """Load Ranker from available models."""
    def __init__(
        self, model_id: str = settings.DEFAULT_MODEL
    ) -> None:
        self.model_id = model_id
        model_file = settings.MODEL_MAP.get(self.model_id)
        if model_file is None:
            raise LookupError(f"{self.model_id!r} model not available.")
        self.instance = ort.InferenceSession(
            settings.get_model_path(model_id=self.model_id) / model_file
        )


class Tokenizer:
    """Load Tokenizer from available models."""
    def __init__(
        self, model_id: str = settings.DEFAULT_MODEL, max_length: int = 512
    ) -> None:
        self.model_id = model_id
        self.model_dir = settings.get_model_path(model_id=self.model_id) 
        self.max_length = max_length
        self.instance = self.__load()
    
    def __file_handler(self, filename: str, read_json: bool = True) -> dict[str, Any] | Path:
        """Json file handler. If read_json is true, returns loaded object else path is returned"""
        path = self.model_dir / filename
        if not path.exists():
            raise FileNotFoundError(f"{filename!r} file missing from {self.model_dir!r}")
        if read_json:
            return json.loads(path.read_bytes())
        return path

    def __load_vocab(self, vocab_file: Path):
        """Load vocab file"""
        vocab, ids_to_tokens = OrderedDict(), OrderedDict() 
        with vocab_file.open(encoding="utf-8") as handler:
            tokens = handler.readlines()

        for idx, tok in enumerate(tokens):
            tok = tok.rstrip("\n")
            vocab[tok], ids_to_tokens[idx] = idx, tok

        return vocab, ids_to_tokens

    def __load(self):
        """Load tokenizer"""
        config = self.__file_handler("config.json")
        tokenizer_config = self.__file_handler("tokenizer_config.json")
        tokens_map = self.__file_handler("special_tokens_map.json")

        tokenizer = cast(TokenizerLoader, TokenizerLoader.from_file(str(
            self.__file_handler("tokenizer.json", read_json=False)
        )))
        tokenizer.enable_truncation(max_length=min(tokenizer_config["model_max_length"], self.max_length))
        tokenizer.enable_padding(pad_id=config["pad_token_id"], pad_token=tokenizer_config["pad_token"])

        for token in tokens_map.values():
            if isinstance(token, str):
                tokenizer.add_special_tokens([token])
            elif isinstance(token, dict):
                tokenizer.add_special_tokens([AddedToken(**token)])
        
        vocab_file = self.model_dir / "vocab.txt"
        if vocab_file.exists():
            tokenizer.vocab, tokenizer.ids_to_tokens = self.__load_vocab(vocab_file)

        return tokenizer
    
    
class ReRankPipeline:
    """
    Pipeline for reranking task.

    Example:
    ```python
    from swiftrank import ReRankPipeline
    pipeline = ReRankPipeline(ranker=ranker, tokenizer=tokenizer)
    pipeline.invoke(
        query="<query>", contexts=["<context1>", "<context2>", ...]    
    )
    ```
    """
    def __init__(
        self, 
        ranker: Optional[Ranker] = None, 
        tokenizer: Optional[Tokenizer] = None
    ) -> None:
        """
        Initialize a rerank pipeline
        @param ranker: `Ranker` class instance
        @param tokenizer: `Tokenizer` class instance
        """
        self.ranker = (ranker or Ranker()).instance
        self.tokenizer = (tokenizer or Tokenizer()).instance

    @classmethod
    def from_model_id(cls, __id: str, tk_max_length: int = 512):
        """
        Create Reranker from model ID
        @param __id: Model ID
        @param tk_max_length: Max length for tokenizer
        """
        return cls(
            ranker=Ranker(model_id=__id), 
            tokenizer=Tokenizer(model_id=__id, max_length=tk_max_length)
        )

    def __create_attr_array(self, tokenized, attr: str):
        """Create array of tokenized attribute values."""
        return np.array([getattr(_, attr) for _ in tokenized], dtype=np.int64)

    @overload
    def invoke_with_score(
        self, query: str, contexts: Iterable[str], threshold: Optional[float] = None
    ) -> list[tuple[float, str]]:
        """
        Rerank contexts based on query.
        @param query: The query to use for reranking evaluation.
        @param contexts: The contexts to rerank.
        @param threshold: Get contexts that are equal or higher than threshold value.
        """
    
    @overload
    def invoke_with_score(
        self, query: str, contexts: Iterable[_T], threshold: Optional[float] = None, *, key: Callable[[_T], str]
    ) -> list[tuple[float, _T]]:
        """
        Rerank contexts based on query.
        @param query: The query to use for reranking evaluation.
        @param contexts: The contexts object.
        @param threshold: Get contexts that are equal or higher than threshold value.
        @param key: callback to use for getting fields from contexts object.
        """

    def invoke_with_score(
        self, 
        query: str, 
        contexts: Iterable, 
        threshold: Optional[float] = None, 
        *, 
        key: Callable = None
    ) -> list[tuple]:

        processor = (lambda _:_) if key is None else key
        tokenized = self.tokenizer.encode_batch(
            [(query, processor(context)) for context in contexts])

        onnx_input = {
            "input_ids": self.__create_attr_array(tokenized, 'ids'),
            "attention_mask": self.__create_attr_array(tokenized, 'attention_mask')}
        token_type_ids = self.__create_attr_array(tokenized, 'type_ids')
        use_type_ids = not np.all(token_type_ids == 0)
        if use_type_ids:
            onnx_input = onnx_input | {'token_type_ids': token_type_ids}

        output = self.ranker.run(None, onnx_input)[0]
        scores = list(1 / (1 + np.exp(
            -(output[:, 1] if output.shape[1] > 1 else output.flatten()))))
        combined = sorted(zip(scores, contexts), key=lambda x: x[0], reverse=True)

        if threshold is None:
            return [(float(sc), ctx) for sc, ctx in combined]
        return [(float(sc), ctx) for sc, ctx in combined if sc >= threshold]

    @overload
    def invoke(
        self, query: str, contexts: Iterable[str], threshold: Optional[float] = None
    ) -> list[str]:
        """
        Rerank contexts based on query.
        @param query: The query to use for reranking evaluation.
        @param contexts: The contexts to rerank.
        @param threshold: Get contexts that are equal or higher than threshold value.
        """
    
    @overload
    def invoke(
        self, query: str, contexts: Iterable[_T], threshold: Optional[float] = None, *, key: Callable[[_T], str]
    ) -> list[_T]:
        """
        Rerank contexts based on query.
        @param query: The query to use for reranking evaluation.
        @param contexts: The contexts object.
        @param threshold: Get contexts that are equal or higher than threshold value.
        @param key: callback to use for getting fields from contexts object.
        """

    def invoke(
        self, 
        query: str, 
        contexts: Iterable, 
        threshold: Optional[float] = None, 
        *, 
        key: Callable = None
    ) -> list:

        return [context for _, context in self.invoke_with_score(
            query=query, contexts=contexts, threshold=threshold, key=key)]


import os
import zipfile
import requests
from pathlib import Path

from tqdm import tqdm

DEFAULT_CACHE_DIR = Path(os.getenv(
    "SWIFTRANK_CACHE", 
    default=Path("~").expanduser() / ".cache" / "swiftrank"
))
DEFAULT_CACHE_DIR.mkdir(parents=True, exist_ok=True)

MODEL_MAP = {
    "ms-marco-TinyBERT-L-2-v2": "flashrank-TinyBERT-L-2-v2.onnx",
    "ms-marco-MiniLM-L-12-v2": "flashrank-MiniLM-L-12-v2_Q.onnx",
    "ms-marco-MultiBERT-L-12": "flashrank-MultiBERT-L12_Q.onnx",
    "rank-T5-flan": "flashrank-rankt5_Q.onnx",
    "ce-esci-MiniLM-L12-v2": "flashrank-ce-esci-MiniLM-L12-v2_Q.onnx"
}

DEFAULT_MODEL = os.getenv("SWIFTRANK_MODEL", "ms-marco-TinyBERT-L-2-v2")
"""Default Model to use"""

def get_model_path(model_id: str) -> Path:
    model_dir = DEFAULT_CACHE_DIR / model_id
    if model_dir.exists():
        return model_dir 

    local_zip_file = str(DEFAULT_CACHE_DIR / f"{model_id}.zip")
    model_url = f"https://huggingface.co/prithivida/flashrank/resolve/main/{model_id}.zip"

    with requests.get(model_url, stream=True) as r:
        r.raise_for_status()
        total_size = int(r.headers.get('content-length', 0))
        progress_bar = tqdm(
            desc=model_id, total=total_size, unit='iB', unit_scale=True, unit_divisor=1024,
        )
        with open(local_zip_file, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                progress_bar.update(f.write(chunk))
        
        progress_bar.desc = local_zip_file
        progress_bar.close()

    with zipfile.ZipFile(local_zip_file, 'r') as zip_ref:
        zip_ref.extractall(DEFAULT_CACHE_DIR)

    os.remove(local_zip_file)
    return model_dir


from typing import Annotated

from cyclopts import App, Parameter, validators

try:
	from signal import signal, SIGPIPE, SIG_DFL
	signal(SIGPIPE, SIG_DFL)
except ImportError:
	pass

app = App(
    name="swiftrank", help="Rerank contexts provided on stdin.",
)

@app.command(name="process", help="STDIN processor. [ json | jsonl | yaml ]")
def build_processing_parameters(
    *,
    pre: Annotated[str, Parameter(name=('-r', '--pre'),
        help="schema for pre-processing input.", show_default=False)] = '.',
    ctx: Annotated[str, Parameter(name=('-c', '--ctx'),
        help="schema for extracting context.", show_default=False)] = '.',
    post: Annotated[str, Parameter(name=('-p', '--post'),
        help="schema for extracting field after reranking.", show_default=False
    )] = None
):  
    from .utils import cli_object_parser, print_and_exit
    def preprocessor(_input: str):
        if _input.startswith(('{', '[')):
            from orjson import loads, JSONDecodeError
            try:
                return cli_object_parser(loads(_input), pre)
            except JSONDecodeError:
                try:
                    from io import StringIO
                    with StringIO(_input) as handler:
                        return list(map(loads, handler))
                except (JSONDecodeError, Exception):
                    print_and_exit("Input data format not valid.", code=1)
            except Exception:
                print_and_exit("Input data format not valid.", code=1)
        
        import yaml    
        try:
            return cli_object_parser(yaml.safe_load(_input), pre)
        except yaml.MarkedYAMLError:
            return list(yaml.safe_load_all(_input))
        except yaml.YAMLError:
            print_and_exit("Input data format not valid.", code=1)
    
    return {'preprocessor': preprocessor, 'ctx_schema': ctx, 'post_schema': post}


@app.meta.default
def __entry__(
    *tokens: Annotated[str, Parameter(show=False, allow_leading_hyphen=True)], 
    query: Annotated[str, Parameter(
        name=("-q", "--query"), help="query for reranking evaluation.")],
    threshold: Annotated[float, Parameter(
        name=("-t", "--threshold"), help="filter contexts using threshold.", validator=validators.Number(gte=0.0, lte=1.0))] = None,
    first: Annotated[bool, Parameter(
        name=("-f", "--first"), help="get most relevant context.", negative="", show_default=False)] = False,
):
    from .utils import read_stdin, cli_object_parser, print_and_exit
    
    processing_params: dict = {}
    if tokens:
        processing_params = app(tokens=tokens)
        _input = read_stdin()
        if not _input:
            return
        contexts = processing_params['preprocessor'](_input)
    else:
        contexts = read_stdin(readlines=True)

    ctx_schema = processing_params.get('ctx_schema', '.')
    post_schema = processing_params.get('post_schema') or ctx_schema

    if not isinstance(contexts, list):
        print_and_exit(cli_object_parser(contexts, post_schema))

    if not contexts:
        print_and_exit("Nothing to rerank!", code=1)

    from .. import settings
    from ..ranker import ReRankPipeline

    pipeline = ReRankPipeline.from_model_id(settings.DEFAULT_MODEL)
    try:
        reranked = pipeline.invoke(
            query=query, 
            contexts=contexts, 
            threshold=threshold, 
            key=lambda x: cli_object_parser(x, ctx_schema)
        )

        if reranked and first:
            print_and_exit(
                cli_object_parser(reranked[0], post_schema)
            )
            
        for context in reranked:
            print(cli_object_parser(context, post_schema))

    except TypeError:
        print_and_exit(
            'Context processing must result into string.', code=1
        )

@app.meta.command(name="serve", help="Startup a swiftrank server")
def serve(
    *, 
    host: Annotated[str, Parameter(
        name=('--host'), help="Host name")] = '0.0.0.0',
    port: Annotated[int, Parameter(
        name=('--port',), help="Port number.")] = 12345
):
    from .api import _serve
    _serve(host=host, port=port)

if __name__ == "__main__":
    app.meta()